
import re
import html
import json
from unidecode import unidecode
from .tools.scraper_tools import ReviewScraper, clean_xpath_res, _aux_clean_number
from .tools.html_tools import get_html
from .tools.generic_tools import logger

class IGraalScraper(ReviewScraper):
    def __init__(self, url, company=""):
        next_page_xpath = "//head/link[@rel='next']/@href"
        company_name_xpath = "//span[@class='shop-name']//text()"

        super().__init__(url=url,
                        next_page_xpath=next_page_xpath,
                        company_name_xpath=company_name_xpath,
                        use_chrome=True,
                        headless=False,
                        )
        print("A window will open. If a captcha is required, please fill it. \
Also close any popup that may appear.")

        get_html(url)
    
    def _parse_review(self, review_block):
        info = dict()
        info["rating_star"] = review_block["reviewRating"]["ratingValue"]
        info["date"] = review_block["datePublished"][:10]
        info["review"] = review_block["reviewBody"]
        return info

    def clean_review(self, r):
        cleaned_r = r.copy()
        cleaned_r["rating_star_cleaned_hide"] = r["rating_star"]
        cleaned_r["date_year_month_hide"] = cleaned_r["date"][:7]
        cleaned_r["text_cleaned_hide"] = unidecode(html.unescape(f"{cleaned_r['review']}"))
        cleaned_r["text_cleaned_hide"] = cleaned_r["text_cleaned_hide"].replace("[^a-zA-Z#]", " ").lower()
        return cleaned_r

    def _parse_page(self, page):
        info = list()
        script = page.xpath("//head/script[@type='application/ld+json']//text()")[0]
        reviews = json.loads(script)["review"]
        for r in reviews:
            res = self._parse_review(r)
            info.append(self.clean_review(res))
        return info

    def scrap_n_reviews(self):
        page_html = self._get_html(self.url)
        script = page_html.xpath("//head/script[@type='application/ld+json']//text()")[0]
        info = json.loads(script)
        self.n_reviews = info["aggregateRating"]["ratingCount"]
    
    def get_n_pages(self):
        n_pages = self.n_reviews // 50 + 1 # 50 reviews per page
        # Not optimised, it's the max # of pages, but often less
        return n_pages

    def is_last_page(self, page):
        return len(page.xpath(self.next_page_xpath)) == 0


def scrap_reviews_igraal(tss, url):
    # tss = IGraalScraper(url)
    info = tss.scrap_website()
    return info



import sys

def singleton(func):
    instances = {}
    def wrapper(force_new=False, *args, **kwargs):
        if force_new or func not in instances:
            instances[func] = func(*args, **kwargs)
        return instances[func]
    return wrapper

def logger(msg):
    sys.stdout.write('\r')
    # sys.stdout.flush()
    sys.stdout.write(msg)


def clean_xpath_res(l):
    return "\n".join(l).strip()

def _aux_clean_number(number_text):
    return int(number_text[0])



import re
import html
import json
from unidecode import unidecode
from .tools.scraper_tools import ReviewScraper, clean_xpath_res, _aux_clean_number


class TrustedshopsScraper(ReviewScraper):
    def __init__(self, url, company=""):
        page_argument = "page"
        company_name_xpath = "//span[@class='shop-name']//text()"

        super().__init__(url=url,
                        page_argument=page_argument,
                        company_name_xpath=company_name_xpath,
                        use_chrome=True,
                        )
    
    def _parse_review(self, review_block):
        info = dict()
        info["rating_star"] = review_block["reviewRating"]["ratingValue"]
        info["date_full_hide"] = review_block["datePublished"]
        info["date"] = review_block["datePublished"][:10]
        info["review"] = review_block["reviewBody"]
        info["language"] = review_block["inLanguage"]
        return info

    def clean_review(self, r):
        cleaned_r = r.copy()
        cleaned_r["rating_star_cleaned_hide"] = r["rating_star"]
        cleaned_r["date_year_month_hide"] = cleaned_r["date"][:7]
        cleaned_r["text_cleaned_hide"] = unidecode(html.unescape(f"{cleaned_r['review']}"))
        cleaned_r["text_cleaned_hide"] = cleaned_r["text_cleaned_hide"].replace("[^a-zA-Z#]", " ").lower()
        return cleaned_r

    def _parse_page(self, page):
        info = list()
        script = page.xpath("//head/script[@type='application/ld+json']//text()")[0]
        reviews = json.loads(script)["review"]
        for r in reviews:
            res = self._parse_review(r)
            info.append(self.clean_review(res))
        return info

    def scrap_n_reviews(self):
        page_html = self._get_html(self.url)
        xpath_n_reviews = "//h2[@class='summary-info']/span[1]//text()"
        n_reviews_str = re.sub("[^0-9]", "", page_html.xpath(xpath_n_reviews)[0])
        self.n_reviews = int(n_reviews_str.replace(" ", "").replace(".", ""))
    
    def get_n_pages(self):
        n_pages = self.n_reviews // 20 + 1 # 20 reviews per page
        # Not optimised, it's the max # of pages, but often less
        return n_pages

    def is_last_page(self, page):
        return page.xpath("//div[@page-index='next']/@style") == ["display: none;"]


def scrap_reviews_trustedshops(url):
    tss = TrustedshopsScraper(url)
    info = tss.scrap_website()
    return info



from lxml import html
import requests
from urllib.parse import (
        urljoin, urlencode, unquote, urlparse, parse_qsl, ParseResult
    )
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from .generic_tools import singleton, clean_xpath_res


def current_page(url, page_argument):
    parsed_get_args = dict(parse_qsl(urlparse(unquote(url)).query))
    return int(parsed_get_args.get(page_argument, 1))

def next_url_page_argument(url, page_argument):
    # Unquoting URL first so we don't loose existing args
    url = unquote(url)
    # Extracting url info
    parsed_url = urlparse(url)
    # Extracting URL arguments from parsed URL
    get_args = parsed_url.query
    # Converting URL arguments to dict
    parsed_get_args = dict(parse_qsl(get_args))
    parsed_get_args[page_argument] = int(parsed_get_args.get(page_argument, 1)) + 1
    # Converting URL argument to proper query string
    encoded_get_args = urlencode(parsed_get_args, doseq=True)
    # Creating new parsed result object based on provided with new
    # URL arguments. Same thing happens inside of urlparse.
    new_url = ParseResult(
        parsed_url.scheme, parsed_url.netloc, parsed_url.path,
        parsed_url.params, encoded_get_args, parsed_url.fragment
    ).geturl()
    return new_url

def next_url_xpath(url, page_html, next_page_xpath):
    return urljoin(url, clean_xpath_res(page_html.xpath(next_page_xpath)))


def next_url(url, page_argument=None, page_html=None, next_page_xpath=None):
    if page_argument is None:
        next_link = next_url_xpath(url, page_html, next_page_xpath)
    else:
        next_link = next_url_page_argument(url, page_argument)
    return next_link


@singleton
def get_driver(headless=True):
    chrome_options = Options()
    #chrome_options.add_argument("--disable-extensions")
    #chrome_options.add_argument("--disable-gpu")
    #chrome_options.add_argument("--no-sandbox") # linux only
    if headless:
        chrome_options.add_argument("--headless")
    
    driver = webdriver.Chrome(options=chrome_options)

    return driver


def get_html(url, use_chrome=False, headless=True, force_new=False):
    if use_chrome:
        driver = get_driver(force_new=force_new, headless=headless)
        driver.get(url)
        page_content = driver.page_source
    else:
        page_content = requests.get(url).content

    page_html = html.fromstring(page_content)

    return page_html




import sys
from lxml import html
import requests
from unidecode import unidecode
from time import sleep
from urllib.parse import (
        urljoin, urlencode, unquote, urlparse, parse_qsl, ParseResult
    )
from .generic_tools import logger, clean_xpath_res, _aux_clean_number
from .html_tools import get_html, current_page, next_url


class ReviewScraper:
    def __init__(self, url, page_argument=None, next_page_xpath="", 
                company="", company_name_xpath="",
                use_chrome=False, headless=True, force_new=False):
        self.url = url
        self.n_reviews = None
        self.reviews = []
        self.page_argument = page_argument
        self.next_page_xpath = next_page_xpath
        self.company = company
        self.company_name_xpath = company_name_xpath
        self.use_chrome = use_chrome
        self.headless = headless
        self.force_new = force_new
        self.company = self.get_company_name()

    def get_company_name(self):
        if self.company:
            return self.company
        if self.company_name_xpath:
            page_html = self._get_html(self.url)
            return clean_xpath_res(page_html.xpath(self.company_name_xpath))
        return ""
    
    def _get_html(self, url):
        return get_html(url, 
                        use_chrome=self.use_chrome, 
                        headless=self.headless, 
                        force_new=self.force_new)
    
    def _parse_review(self, review_block):
        return dict()

    def clean_review(self, review):
        return dict()

    def _parse_page(self, page):
        return []

    def scrap_n_reviews(self):
        return 0
    
    def get_n_pages(self):
        return 0

    def is_last_page(self, page):
        return True

    def _pause(self, n):
        if n % 1000 == 0:
            time_sleep = 25
        elif n % 100 == 0:
            time_sleep = 7
        elif n % 10 == 0:
            time_sleep = 1
        else:
            time_sleep = 0

        if self.n_reviews > 10000:
            time_sleep = 3 * time_sleep + 1
        elif self.n_reviews > 1000:
            time_sleep *= 2

        if time_sleep > 5:
            logger(f"Sleeping for {time_sleep}sec to avoid being blocked")
        if time_sleep:
            sleep(time_sleep)
    
    def scrap_reviews(self):
        print("Starting to scrap...")
        
        n_pages = self.get_n_pages()
        url = self.url
        n_start = current_page(url, self.page_argument)
        for n in range(n_start, n_pages + 1):
            page_html = self._get_html(url)
            page_info = self._parse_page(page_html)
            if len(page_info) == 0:
                break
            for i in page_info:
                i["page"] = n
                i["company"] = self.company
            self.reviews += page_info

            n_done = (n - n_start + 1)
            n_total = (n_pages - n_start + 1)
            percent_done = int( n_done / n_total * 100)

            display_msg = f"Done {n_done}/{n_total} ({percent_done}%): {url}"
            
            logger(display_msg)

            if self.is_last_page(page_html):
                break
            self._pause(n - n_start + 1)
            url = next_url(url, 
                            page_argument=self.page_argument,
                            page_html=page_html,
                            next_page_xpath=self.next_page_xpath)
    
    def scrap_website(self):
        self.scrap_n_reviews()
        self.scrap_reviews()
        res = {
        	"url": self.url,
        	"n_reviews": self.n_reviews,
        	"reviews": self.reviews
        }
        return res


import pandas as pd

def avg_rating_month_data(df_reviews):
    """
    Needs two columns: 'date_year_month_hide' and 'rating_star_cleaned_hide'
    """
    df = df_reviews.copy()
    df.date_year_month_hide = pd.to_datetime(df.date_year_month_hide)
    df.sort_values("date_year_month_hide", inplace=True)
    df["sum_rating"] =  df["rating_star_cleaned_hide"].cumsum()
    df["n_rating"] =  1
    df["n_rating"] =  df["n_rating"].cumsum()
    df["avg_rating_uptodate"] = df["sum_rating"] / df["n_rating"]
    agg_func = {
        "avg_rating_uptodate": "last",
        "rating_star_cleaned_hide": "mean"
    }
    df_months = df.groupby(["date_year_month_hide"]).agg(agg_func).stack().reset_index()
    df_rating_star = df_months[df_months.level_1 == "rating_star_cleaned_hide"].set_index(
                            ["date_year_month_hide", "level_1"]
                        ).unstack(
                            fill_value=0
                        ).asfreq(
                            'MS', fill_value=0
                        ).stack().sort_index(level=1).reset_index()

    df_avg = df_months[df_months.level_1 == "avg_rating_uptodate"].set_index(
                            ["date_year_month_hide", "level_1"]
                        ).unstack(
                            fill_value=0
                        ).asfreq(
                            'MS', method="ffill"
                        ).stack().sort_index(level=1).reset_index()

    df_months = pd.concat([df_rating_star, df_avg])

    df_months = pd.crosstab(
            df_months["level_1"],
            df_months["date_year_month_hide"],
            df_months[0],
            aggfunc="sum"
    )

    df_months.index = ["Average Rating up to Date", "Monthly Average Rating"]
    df_months.columns.name = "Date"

    return df_months

def n_rating_month(df_reviews):
    """
    Needs two columns: 'date_year_month_hide' and 'rating_star_cleaned_hide', 'text_cleaned_hide'
    """
    df = df_reviews.copy()
    df.date_year_month_hide = pd.to_datetime(df.date_year_month_hide)
    df_n_rating = df.groupby(
            ["date_year_month_hide", "rating_star_cleaned_hide"]
        )["text_cleaned_hide"].count().reset_index()
    df_n_rating = df_n_rating.set_index(
                                ["date_year_month_hide", "rating_star_cleaned_hide"]
                            ).unstack(
                                fill_value=0
                            ).asfreq(
                                'MS', fill_value=0
                            ).stack().sort_index(level=1).reset_index()

    df_n_rating = pd.crosstab(
            df_n_rating["rating_star_cleaned_hide"],
            df_n_rating["date_year_month_hide"],
            df_n_rating["text_cleaned_hide"],
            aggfunc="sum")

    df_n_rating.columns.name = "Date"
    df_n_rating.index.name = "Rating"

    return df_n_rating

def write_dfs(dict_dfs, filename):

    writer = pd.ExcelWriter(filename, engine='xlsxwriter')
    for sheetname, data in dict_dfs.items():
        data.to_excel(writer, sheet_name=sheetname)
    writer.save()

    print(f"{filename} saved")
    
    
def get_analysis(df, filename=""):
    df_avg_rating = avg_rating_month_data(df)
    df_n_rating = n_rating_month(df)

    dict_dfs = {
        "Reviews": df,
        "Average Rating": df_avg_rating,
        "N Rating": df_n_rating,
    }
    
    if filename:
        write_dfs(dict_dfs, filename)
    
    return dict_dfs


from .tools.scraper_tools import ReviewScraper, clean_xpath_res, _aux_clean_number
import re


class AvisVerifiesScraper(ReviewScraper):
    def __init__(self, url, company=""):
        page_argument = "p"
        company_name_xpath = "//span[@class='label-title']//span[@itemprop='name']//text()"
        super().__init__(url=url,
                        page_argument=page_argument,
                        company=company,
                        company_name_xpath=company_name_xpath)
    
    def _parse_review(self, review_block):
        info = dict()
        xpath_rating = ".//span[@itemprop='ratingValue']//text()"
        info["rating_star"] = clean_xpath_res(review_block.xpath(xpath_rating))
        xpath_date = ".//meta[@itemprop='datePublished']/@content"
        info["date"] = str(review_block.xpath(xpath_date)[0])
        xpath_review = ".//div[@itemprop='reviewBody']//text()"
        info["review"] = clean_xpath_res(review_block.xpath(xpath_review))
        xpath_details = ".//div[@class='details suite']//text()"
        info["details_hide"] = clean_xpath_res(review_block.xpath(xpath_details))
        xpath_answer = ".//div[@class='answers']/div[@class='merchant']//div[@class='text']//text()"
        info["answer"] = clean_xpath_res(review_block.xpath(xpath_answer))
        info["company_has_answered"] = len(info["answer"]) > 0
        return info

    def clean_review(self, r):
        cleaned_r = r.copy()
        cleaned_r["rating_star_cleaned_hide"] = _aux_clean_number(cleaned_r["rating_star"])
        cleaned_r["date_year_month_hide"] = f"{cleaned_r['date'][-4:]}-{cleaned_r['date'][3:5]}"
        cleaned_r["text_cleaned_hide"] = cleaned_r["review"].replace("[^a-zA-Z#]", " ").lower()

        return cleaned_r
        
    def _parse_page(self, page):
        info = list()
        reviews = page.xpath("//div[@class='review row']")
        for r in reviews:
            res = self._parse_review(r)
            info.append(self.clean_review(res))
        return info

    def scrap_n_reviews(self):
        page_html = self._get_html(self.url)
        xpath_n_reviews = "//span[@class='count']//text()"
        n_reviews_str = re.sub("[^0-9]", "", page_html.xpath(xpath_n_reviews)[0])
        self.n_reviews = int(n_reviews_str.replace(" ", ""))
    
    def get_n_pages(self):
        page_html = self._get_html(self.url)
        xpath_pages = "//select[@class='selectPage']/option"
        list_pages = page_html.xpath(xpath_pages)
        n_pages = len(list_pages)
        return n_pages

    def is_last_page(self, page):
        return len(page.xpath("//a[@rel='next']/@href")) == 0
    
    
def scrap_reviews_avisverifies(url):
    avs = AvisVerifiesScraper(url)
    info = avs.scrap_website()
    return info




from .tools.scraper_tools import ReviewScraper, clean_xpath_res, _aux_clean_number
import json
import re


class TrustpilotScraper(ReviewScraper):
    def __init__(self, url, company=""):
        # page_argument = "page"
        next_page_xpath = "//a[@data-page-number='next-page']/@href"
        company_name_xpath = "//span[@class='multi-size-header__big']//text()"
        super().__init__(url=url,
                        next_page_xpath=next_page_xpath,
                        company=company,
                        company_name_xpath=company_name_xpath)
    
    def _parse_review(self, review_block):
        info = dict()
        info["company"] = self.company
        xpath_id = ".//h2/a/@href"
        info["review_id"] = clean_xpath_res(review_block.xpath(xpath_id))
        xpath_rating = ".//div[contains(@class, 'reviewHeader']/@data-service-review-rating"
        info["rating_star"] = clean_xpath_res(review_block.xpath(xpath_rating))
        # xpath_date = ".//div[@class='review-content-header__dates']/script/text()"
        # info["date"] = json.loads(review_block.xpath(xpath_date)[0])["publishedDate"]
        # xpath_title = ".//h2//text()"
        # info["title"] = clean_xpath_res(review_block.xpath(xpath_title))
        # xpath_review = ".//p[contains(@class, 'tyography_body']//text()"
        # info["review"] = clean_xpath_res(review_block.xpath(xpath_review))
        xpath_customer_location = ".//div[contains(@class, 'consumer-information__location')]//text()"
        info["customer_location"] = clean_xpath_res(review_block.xpath(xpath_customer_location))
        xpath_n_reviews = ".//div[@class='consumer-information__review-count']//text()"
        info["n_reviews_customer_hide"] = clean_xpath_res(review_block.xpath(xpath_n_reviews))
        xpath_is_verified_info = ".//div[contains(@class , 'review-content-header__review-labels')]/script/text()"
        is_verified_info = review_block.xpath(xpath_is_verified_info)
        if len(is_verified_info) > 0:
            is_verified_info = json.loads(is_verified_info[0])
        else:
            is_verified_info = dict()
        info["is_verified"] = is_verified_info.get('isVerified', False)
        info["verification_source_hide"] = is_verified_info.get('verificationSource', "")
        info["review_source_hide"] = is_verified_info.get('reviewSourceName', "")
        info["verification_level_hide"] = is_verified_info.get('verificationLevel', "")
        xpath_reply_date = ".//script[@data-initial-state='review-reply-dates']//text()"
        reply_date_info = clean_xpath_res(review_block.xpath(xpath_reply_date))
        if reply_date_info:
            reply_date_info = json.loads(reply_date_info)
            info["reply_date_hide"] = reply_date_info["publishedDate"]
        else:
            info["reply_date_hide"] = ""
        xpath_reply_content = ".//div[@class='brand-company-reply__content']//text()"
        info["reply_content_hide"] = clean_xpath_res(review_block.xpath(xpath_reply_content))
        return info

    def _parse_review_script(self, script):
        info = dict()
        info["reviewId"] = script.get("id", "")
        info["language"] = script.get("language", "")
        info["company"] = self.company
        info["title"] = script.get("title", "")
        info["review"] = script.get("text", "")
        info["rating_star"] = script.get("rating", "")
        try:
            info["date"] = script.get("dates", dict()).get("publishedDate", "")
        except:
            pass

        try:
            consumer_info = script.get("consumer", dict())
            info["customer_location"] = consumer_info.get("countryCode", "")
            info["n_reviews_customer_hide"] = consumer_info.get("numberOfReviews", "")
        except:
            pass

        try:
            verified_info = script.get("labels", dict()).get("verification", dict())
            info["is_verified"] = verified_info.get("isVerified", "")
            info["verification_source_hide"] = verified_info.get("verificationSource", "")
            info["review_source_hide"] = verified_info.get("reviewSourceName", "")
            info["verification_level_hide"] = verified_info.get("verificationLevel", "")
        except:
            pass

        try:
            reply_info = script.get("reply", dict())
            info["reply_text_hide"] = reply_info.get("message", "")
            info["reply_date_hide"] = reply_info.get("publishedDate", "")
        except:
            pass

        return info

    def clean_review(self, r):
        cleaned_r = r.copy()
        # cleaned_r["rating_star_cleaned_hide"] = _aux_clean_number(r["rating_star"])
        # cleaned_r["n_reviews_customer_cleaned_hide"] = _aux_clean_number(r["n_reviews_customer_hide"])
        cleaned_r["date_year_month_hide"] = cleaned_r["date"][:7]
        cleaned_r["reply_date_year_month_hide"] = cleaned_r["reply_date_hide"][:7]
        cleaned_r["text_cleaned_hide"] = f"{cleaned_r['title']} {cleaned_r['review']}"
        cleaned_r["text_cleaned_hide"] = cleaned_r["text_cleaned_hide"].replace("[^a-zA-Z#]", " ").lower()
        return cleaned_r

    # def _parse_page(self, page):
    # """ Old version"""
    #     info = list()
    #     reviews = page.xpath("//div[contains(@class , 'review-card')]")
    #     reviews_script = page.xpath("//script[@type='application/json']//text()")
    #     if reviews_script:
    #         reviews_script = reviews_script[0]
    #         reviews_info = json.loads(reviews_script)["props"]["pageProps"]["reviews"]
    #     else:
    #         reviews_info = [dict() for r in reviews]
    #     for r, r_info in zip(reviews, reviews_info):
    #         res = self._parse_review(r)
    #         res_info = self._parse_review_script(r_info)
    #         if res["rating_star"] != '#rating/desc/star0#':
    #             res["language"] = r_info.get("inLanguage", "Unknown")
    #             info.append(self.clean_review(res))
    #     return info

    def _parse_page(self, page):
        info = list()
        reviews_script = page.xpath("//script[@type='application/json']//text()")
        reviews_info = []
        if reviews_script:
            reviews_script = reviews_script[0]
            reviews_info = json.loads(reviews_script)["props"]["pageProps"]["reviews"]
        for review in reviews_info:
            res = self._parse_review_script(review)
            info.append(self.clean_review(res))
        return info


    def scrap_n_reviews(self):
        page_html = self._get_html(self.url)
        # xpath_n_reviews = "//h2[@class='header--inline']//text()"
        # n_reviews_str = re.sub("[^0-9]", "", page_html.xpath(xpath_n_reviews)[0])
        # self.n_reviews = int(n_reviews_str.replace(" ", "").replace(".", ""))
        n_xpath = "//script[@id='__NEXT_DATA__']//text()"
        script = page_html.xpath(n_xpath)[0]
        json_script = json.loads(script)
        self.n_reviews = json_script["props"]["pageProps"]["businessUnit"]["numberOfReviews"]


    def get_n_pages(self):
        n_pages = self.n_reviews // 20 + 1 # 20 reviews per page
        # Not optimised, it's the max # of pages, but often less
        return n_pages

    def is_last_page(self, page):
        return len(page.xpath("//a[@dname='pagination-button-next']")) == 0


def scrap_reviews_trustpilot(url):
    tps = TrustpilotScraper(url)
    info = tps.scrap_website()
    return info



from .scrap_avisverifies import scrap_reviews_avisverifies
from .scrap_trustpilot import scrap_reviews_trustpilot
from .scrap_trustedshops import scrap_reviews_trustedshops, TrustedshopsScraper
from .scrap_igraal import scrap_reviews_igraal, IGraalScraper